from __future__ import annotations
import inspect
import logging
from enum import Enum, unique
from typing import TYPE_CHECKING, Generator


if TYPE_CHECKING:
    from .ai_care import AICare


logger = logging.getLogger("ai_care")


@unique
class Choice(Enum):
    ERROR = '00'
    STAY_SILENT = '01'
    SPEAK_NOW = '02'
    SPEAK_AFTER = '03'
    DETECT_ENV = '04'
    RELEASE_DETECTOR = '05'
    ASK_LATER = '06'
    CYCLIC_DETECTION = '07'


def _ability(description: str):
    def decorator(func):
        func._ability_ = True
        func._ability_description_ = description
        if not hasattr(func, "_ability_parameters_"):
            func._ability_parameters_ = []
        return func
    return decorator

def _ability_parameter(
    *,
    name: str,
    description: str,
    default_value: str = "",
    param_type: str = "string",
    required: bool = True
):
    """
    Decorator for ability parameters.

    Args:
        name: The name of the ability parameter
        description: The description of the ability parameter
        default_value: The default value of the ability parameter
        type: The type of the ability parameter
        required: Whether the ability parameter is required
    """

    def decorator(func):
        if not hasattr(func, "_ability_parameters_"):
            func._ability_parameters_ = []

        func._ability_parameters_.append(
            {
                "name": name,
                "description": description,
                "default_value": default_value,
                "param_type": param_type,
                "required": required,
            }
        )
        return func

    return decorator

def _auto_depth(depth_param_name: str = "_depth_left"):
    def decorator(func):
        func._auto_depth_ = True
        func._depth_param_name_ = depth_param_name
        return func
    return decorator
    

class Ability:
    def __init__(self, ai_care: AICare):
        self.ai_care = ai_care
        self.abilities: dict = {}
        self._register_abilities()

    def _register_abilities(self) -> None:
        for name, method in inspect.getmembers(self, predicate=inspect.ismethod):
            if getattr(method, '_ability_', False):
                self.abilities[name] = method

    @_ability(
        description="Remain silent.",
    )
    def stay_silent(self) -> None:
        return
    
    @_ability(
        description="Speak to the user right now.",
    )
    @_ability_parameter(
        name="content",
        description="The content you want to say.",
    )
    def speak_now(self, content: str | Generator[str, None, None]) -> None:
        self.ai_care.to_user_method(content)

    @_ability(
        description=(
            "Set a message to be delivered to the user after a certain period of time. "
            "This choice means you decide to observe for a duration, and if the user does "
            "not speak to you during this time, you will then convey the message content you have set. "
            "However, if the user speaks to you within this time frame, the operation will automatically "
            "be cancelled. This option is recommended."
        )
    )
    @_ability_parameter(
        name="delay",
        description="Set how long until your message is sent. Unit in seconds.",
    )
    @_ability_parameter(
        name="message",
        description="Set what you want to say to the user after a certain period of time.",
    )
    def speak_after(self, delay: float | int, message: str) -> None:
        if self.ai_care._stream_mode is True:
            message_wrap = (string for string in [message])
        else:
            message_wrap = message
        self.ai_care.set_timer(interval=delay, function=self.ai_care.to_user_method, args=(message_wrap,))
    
    @_ability(
        description=(
            "Detect environmental conditions. This choice means that you decide to first obtain the results "
            "from the sensors, observe the environmental situation, and then decide what to choose based on "
            "this information. You can only choose which sensors to use from the list of available sensors."
        ),
    )
    @_ability_parameter(
        name="delay",
        description="Set how long to wait before using sensors to obtain readings. Unit in seconds.",
    )
    @_ability_parameter(
        name="sensors",
        description="The list of names of the sensors to be used.",
        param_type="list[str]",
    )
    @_auto_depth(depth_param_name="_depth_left")
    def detect_env(self, delay: float | int, sensors: list[str], _depth_left: int) -> None:
        def detect_env_callback(sensors_list: list[str]):
            sensor_data = {}
            for sensor in sensors_list:
                data = self.ai_care.get_sensor_data(sensor)
                sensor_data[sensor] = data
            self.ai_care.ask(
                messages_list=[
                    {
                        "role": "ai_care",
                        "content": f"The results of the sensor are as follows: {str(sensor_data)}.",
                    },
                ],
                depth_left = _depth_left - 1,
            )
        not_existed_sensors_set = set(sensors) - set(self.ai_care.sensors)
        if not_existed_sensors_set:
            self.ai_care.ask(
                messages_list=[
                    {
                        "role": "ai_care",
                        "content": f"There are no {str(not_existed_sensors_set)} sensor. Please use the correct sensor name.",
                    }
                ],
                depth_left = _depth_left - 1,
            )
            return
        self.ai_care.set_timer(
            interval=delay,
            function=detect_env_callback,
            args=(sensors,),
        )

    @_ability(
        description=(
            "Release detectors. This option means you forego making an active choice and instead release "
            "some detectors, which will automatically report back to you and ask for your decision when "
            "specific conditions are met. You can only choose which detectors to use from the list of "
            "available detectors."
        ),
    )
    @_ability_parameter(
        name="delay",
        description="Set the time in seconds before releasing the detectors.",
    )
    @_ability_parameter(
        name="detectors",
        description="The list of names of the detectors to be released.",
    )
    @_auto_depth(depth_param_name="_depth_left")
    def release_detector(self, delay: int | float, detectors: list[str], _depth_left: int) -> None:
        def release_detector_callback(detectors_list: list[str]):
            self.ai_care.release_detector(detectors_list)
        not_existed_detectors_set = set(detectors) - set(self.ai_care.detectors)
        if not_existed_detectors_set:
            self.ai_care.ask(
                messages_list=[
                    {
                        "role": "ai_care",
                        "content": f"There are no {str(not_existed_detectors_set)} detector. Please use the correct detector name.",
                    }
                ],
                depth_left = _depth_left - 1,
            )
            return
        self.ai_care.set_timer(interval=delay, function=release_detector_callback, args=(detectors,))

    @_ability(
        description=(
            "Ask again after a while. This choice means that you do not want to make a decision right now, "
            "and you would like Aicarey to ask you again later."
        ),
    )
    @_ability_parameter(
        name="delay",
        description="Set how long to wait before asking you again, in seconds.",
    )
    @_auto_depth(depth_param_name="_depth_left")
    def ask_later(self, delay: int | float, _depth_left: int) -> None:
        if self.ai_care._ask_later_count_left <= 0:
            return
        self.ai_care._ask_later_count_left -= 1
        self.ai_care.set_timer(
            interval=delay,
            function=self.ai_care.ask,
            kwargs={
                "messages_list": [
                    {
                        "role": "ai_care",
                        "content": (
                            f"{delay} senconds ago, you asked me to inquire later, "
                            f"and now {delay} seconds have passed. Please make a choice."
                        )
                    }
                ],
                "depth_left": _depth_left - 1,
            },
        )

    @_ability(
        description=(
            "Cycle release of detectors. This option means that you forgo making an active choice "
            "and instead continuously release some detectors, which will automatically report back "
            "to you and ask for your decision when specific conditions are met. The detectors you "
            "select will be released periodically at set time intervals until the next conversation "
            "is initiated. All chosen detectors will be released in each cycle."
        ),
    )
    @_ability_parameter(
        name="interval",
        description="Set the time interval between each release of the detectors. Unit in seconds",
    )
    @_ability_parameter(
        name="detectors",
        description="The list of names of the detectors to be released.",
    )
    @_auto_depth(depth_param_name="_depth_left")
    def cyclic_detection(self, interval: int | float, detectors: list[str], _depth_left) -> None:
        not_existed_detectors_set = set(detectors) - set(self.ai_care.detectors)
        if not_existed_detectors_set:
            self.ai_care.ask(
                messages_list=[
                    {
                        "role": "ai_care",
                        "content": f"There are no {str(not_existed_detectors_set)} detector. Please use the correct detector name.",
                    }
                ],
                depth_left = _depth_left - 1,
            )
            return
        def repeat_interval(interval: float):
            while True:
                yield interval
        self.ai_care.set_cyclic_detection(
            detectors=detectors,
            interval_gen=repeat_interval(float(interval)),
            cancel_after_trigger_ask=True
        )



import pytest
import textwrap
from enum import Enum
from unittest.mock import Mock, patch

from ai_care import AICare
from ai_care.render_prompt import _render_ability_description, render_basic_prompt


@pytest.fixture
def ai_care():
    ai_care = AICare()
    yield ai_care
    ai_care.clear_timer(clear_preserved=True, default_task_num_authority_external="Highest")

def test_render_ability_description():
    
    # Setup
    class MockChoice(Enum):
        STAY_SILENT = "01"
        SPEAK_NOW = "02"
        WITH_PARAMETER = "03"

    ability1, ability2, ability3 = Mock(), Mock(), Mock()
    
    ability1.__name__ = "stay_silent"
    ability1._ability_description_ = "ability stay silent have no parameter"
    ability1._ability_parameters_ = []
    
    ability2.__name__ = "speak_now"
    ability2._ability_description_ = "ability speak now is not a form of parameter calling; it derectly provides content"
    ability2._ability_parameters_ = []

    ability3.__name__ = "with_parameter"
    ability3._ability_description_ = "ability with parameter description"
    ability3._ability_parameters_ = [
        {
            "name": "param_1",
            "param_type": "list[str]",
            "description": "param_1 description",
            "required": True,
            "default_value": [],
        },
        {
            "name": "param_2",
            "param_type": "string",
            "description": "param_2 description",
            "required": False,
            "default_value": "default value",
        },
    ]
    
    # Action
    with patch('ai_care.render_prompt.Choice', MockChoice):
        prompt1 = _render_ability_description(ability1)
        prompt2 = _render_ability_description(ability2)
        prompt3 = _render_ability_description(ability3)
    
    # Assert
    assert prompt1 == textwrap.dedent(
        """
        Choice code: 01
        Choice description: ability stay silent have no parameter
        Choice content: Empty

        """
    )
    assert prompt2 == textwrap.dedent(
        """
        Choice code: 02
        Choice description: ability speak now is not a form of parameter calling; it derectly provides content
        Choice content: The content you want to say

        """
    )
    assert prompt3 == textwrap.dedent(
        """
        Choice code: 03
        Choice description: ability with parameter description
        Choice content: Provide the values of the following parameters in JSON format
            Parameter name: param_1
            Parameter description: param_1 description
            Parameter type: list[str]
            This parameter must be provided
            
            Parameter name: param_2
            Parameter description: param_2 description
            Parameter type: string
            This parameter is optional, with a default value: default value

        """
    )

def test_render_basic_prompt(ai_care: AICare):
    prompt = render_basic_prompt(ai_care)
    assert isinstance(prompt, str)



import pytest
from typing import Generator

from ai_care import AICare
from ai_care.parse_response import parse_response, _extract_info


@pytest.fixture
def ai_care():
    ai_care = AICare()
    yield ai_care
    ai_care.clear_timer(clear_preserved=True, default_task_num_authority_external="Highest")

def test_parse_response(ai_care: AICare):
    # Setup
    response1 = "AA000101:stay silent"
    response2 = "AA000202:speak content"
    response3 = "AA000303:content"
    response_stream1 = (x for x in ["AA", "0001", "01", ":stay ", "si", "lent"])
    response_stream2 = (x for x in ["AA", "0002", "0", "2:spe", "ak ", "now"])
    response_stream3 = (x for x in ["AA", "0002", "02", ":spe", "ak ", "now"])
    response_stream4 = (x for x in ["AA", "0003", "03", ":parameter ", "content"])
    
    # Action
    choice_code1, content1 = parse_response(ai_care=ai_care, response=response1)
    choice_code2, content2 = parse_response(ai_care=ai_care, response=response2)
    choice_code3, content3 = parse_response(ai_care=ai_care, response=response3)
    choice_code_stream1, content_stream1 = parse_response(ai_care=ai_care, response=response_stream1)
    choice_code_stream2, content_stream2 = parse_response(ai_care=ai_care, response=response_stream2)
    choice_code_stream3, content_stream3 = parse_response(ai_care=ai_care, response=response_stream3)
    choice_code_stream4, content_stream4 = parse_response(ai_care=ai_care, response=response_stream4)
    
    # Assert
    assert choice_code1 == "01"
    assert content1 == "stay silent"
    assert choice_code2 == "02"
    assert content2 == "speak content"
    assert choice_code3 == "03"
    assert content3 == "content"
    assert choice_code_stream1 == "01"
    assert content_stream1 == "stay silent"
    assert choice_code_stream2 == "02"
    assert isinstance(content_stream2, Generator)
    assert ''.join(x for x in content_stream2) == "speak now"
    assert choice_code_stream3 == "02"
    assert ''.join(x for x in content_stream3) == "speak now"
    assert choice_code_stream4 == "03"
    assert content_stream4 == "parameter content"

def test_extract_info():
    # Setup
    case1 = "AA0"
    case2 = "ABCd04"
    case3 = "ABCd4560"
    case4 = "AA000202:content"

    # Action
    code1, content1 = _extract_info(case1)
    code2, content2 = _extract_info(case2)
    code3, content3 = _extract_info(case3)
    code4, content4 = _extract_info(case4)

    # Assert
    assert code1 == ("AA", None, None, None)
    assert content1 == None
    assert code2 == ("AB", "Cd", "04", None)
    assert content2 == None
    assert code3 == ("AB", "Cd", "45", "60")
    assert content3 == None
    assert code4 == ("AA", "00", "02", "02")
    assert content4 == "content"



from __future__ import annotations
import itertools
import logging
import time
import threading
from abc import ABCMeta, abstractmethod
from typing import Callable, Any, Generator, TypedDict, Literal, cast

from .abilities import Ability
from .choice_execute import choice_execute
from .parse_response import parse_response
from .render_prompt import render_basic_prompt


logger = logging.getLogger("ai_care")
ChatContext = Any
ConfigKey = Literal["delay", "ask_later_count_limit", "ask_depth", "n_chat_intervals"]


class AICare:

    def __init__(self) -> None:
        self.timers: dict[int, AICareTimer] = {}
        self.detectors: dict[str, Detector] = {}
        self.sensors: dict[str, dict] = {}
        self.ability: Ability = Ability(self)
        self.chat_context: Any = None
        self._to_llm_method: (
            Callable[[ChatContext, list[AICareContext]], str] |
            Callable[[ChatContext, list[AICareContext]], Generator[str, None, None]]
        ) = self._fake_to_llm_method
        self._to_user_method: (
            Callable[[str], None] |
            Callable[[Generator[str, None, None]], None]
        ) = self._fake_to_user_method
        self.guide: str = ""
        self._unique_id = itertools.count(1)
        self._last_chat_time: float | None = None
        self._chat_intervals: list[float] = []
        self._tags: dict[str, list[Detector]] = {}
        self._config: dict[str, Any] = {"delay": 100, "ask_later_count_limit": 1, "ask_depth": 1, "n_chat_intervals": 20}
        self._ask_later_count_left = self._config["ask_later_count_limit"]
        self._valid_msg_count: int = 0
        self._invalid_msg_count: int = 0
        self._stream_mode: bool = True
        self._ask_context: list[AICareContext] = []
        self._task_num: int = 1
    
    @property
    def health(self) -> float:
        return self._valid_msg_count / (self._valid_msg_count + self._invalid_msg_count)

    def cancel_current_task(self):
        self._task_num += 1

    def reset(self) -> None:
        self._valid_msg_count = 0
        self._invalid_msg_count = 0
        self._last_chat_time = None
        self._chat_intervals = []
        self.cancel_current_task()
        self.clear_timer(clear_preserved=True)

    def set_guide(self, guide: str) -> None:
        """Set guidance information.

        This information can be any general type of information,
        such as descriptions of AI personality, descriptions of codes of conduct, etc.
        This information will be inserted into every round of dialogue between ai_care and the AI.

        Example:
            - You are a very considerate person who cares about others and is willing to initiate conversations.
            - You are someone who does not like to meddle in others' affairs, and you do not talk much unless necessary.
        """
        if not isinstance(guide, str):
            raise TypeError(f"Expected a guide of string type, but received a guide of type {type(guide)}.")
        self.guide = guide

    def _fake_to_llm_method(self, chat_context: ChatContext, to_llm_messages: list[AICareContext]) -> str:
        raise NotImplementedError("'to_llm_method' method has not been implemented yet.")
    
    def _fake_to_user_method(self, to_user_message: str) -> None:
        raise NotImplementedError("'to_user_method' method has not been implemented yet.")

    def register_to_llm_method(
        self,
        to_llm_method: Callable[[ChatContext, list[AICareContext]], str] | Callable[[ChatContext, list[AICareContext]], Generator[str, None, None]],
    ) -> None:
        """Register the method used by AICare to send message to llm."""
        self._to_llm_method = to_llm_method

    def register_to_user_method(
        self,
        to_user_method: Callable[[str], None] | Callable[[Generator[str, None, None]], None],
    ) -> None:
        """Register the method used by AICare to send message to user."""
        self._to_user_method = to_user_method

    def to_llm_method(self, chat_context: ChatContext, messages_list: list[AICareContext]) -> str | Generator[str, None, None]:
        return self._to_llm_method(chat_context, messages_list)

    def to_user_method(self, message: str | Generator[str, None, None]) -> None:
        if self._check_task_validity():
            if isinstance(message, str):
                self._to_user_method = cast(Callable[[str], None], self._to_user_method)
                self._to_user_method(message)
            elif isinstance(message, Generator):
                self._to_user_method = cast(Callable[[Generator[str, None, None]], None], self._to_user_method)
                self._to_user_method(message)
            else:
                assert False

    def _check_task_validity(self) -> bool:
        thread_instance = threading.current_thread()
        if isinstance(thread_instance, AICareTimer) or isinstance(thread_instance, AICareThread):
            _task_num = thread_instance._task_num
        else:
            _task_num = self._task_num
        if _task_num < self._task_num:
            return False
        else:
            return True

    def chat_update(self, chat_context: ChatContext) -> None:
        self._task_num += 1
        time_now = time.monotonic()
        if self._last_chat_time is not None:
            self._insert_chat_interval(time_now - self._last_chat_time)
        self._last_chat_time = time_now
        self.chat_context = chat_context
        self._ask_later_count_left = self._config["ask_later_count_limit"]
        self.clear_timer(clear_preserved=False)
        self._ask_context = []
        self.set_timer(
            interval=self._config["delay"],
            function=self.ask,
            kwargs={
                "messages_list": [
                    {
                        "role": "ai_care",
                        "content": render_basic_prompt(self),
                    }
                ],
            },
            task_num=self._task_num,
        )

    def _insert_chat_interval(self, interval: float) -> None:
        if len(self._chat_intervals) >= self._config["n_chat_intervals"]:
            self._chat_intervals.pop(0)
        self._chat_intervals.append(interval)

    def set_config(self, key: ConfigKey, value: Any) -> None:
        # Valid check.
        if key not in {"delay", "ask_later_count_limit", "ask_depth", "n_chat_intervals"}:
            raise TypeError(f"AICare does not accept {key} as a config.")
        self._config[key] = value

    def set_timer(
        self,
        interval: float | int,
        function: Callable,
        args: tuple | None = None,
        kwargs: dict | None = None,
        *,
        preserve: bool = False,
        daemon: bool = True,
        task_num: int | None = None,
    ) -> int:
        args = args or ()
        kwargs = kwargs or {}
        id = next(self._unique_id)
        timer = AICareTimer(interval=float(interval), function=self._timer_wrap, args=(function, id, *args), kwargs=kwargs)
        timer._task_num = task_num or self._get_task_num()
        timer._preserve_ = preserve
        timer.daemon = daemon
        timer.start()
        self.timers[id] = timer
        return id

    def _get_task_num(self) -> int:
        thread_instance = threading.current_thread()
        task_num = getattr(thread_instance, '_task_num', self._task_num)
        return task_num

    def clear_timer(
        self,
        task_num_authority: int | None = None,
        clear_preserved: bool = True,
        default_task_num_authority_external: Literal["Highest", "Lowest"] = "Highest",
    ) -> None:
        keys = list(self.timers.keys())
        if task_num_authority is None:
            thread_instance = threading.current_thread()
            task_num_authority = getattr(
                thread_instance,
                '_task_num',
                self._task_num if default_task_num_authority_external == "Highest" else 0,
            )
        
        for key in keys:
            timer = self.timers[key]
            if timer._task_num <= task_num_authority and timer._preserve_ <= clear_preserved:
                self.timers.pop(key)
                timer.cancel()

    def timer_cancel(self, id: int) -> None:
        timer = self.timers.pop(id, None)
        if timer is not None:
            timer.cancel()

    def _timer_wrap(self, function: Callable, id: int, *args, **kwargs) -> None:
        function(*args, **kwargs)
        try:
            del self.timers[id]
        except KeyError:
            pass

    def ask(
        self,
        messages_list: list[AICareContext],
        chat_context: ChatContext | None = None,
        depth_left: int | None = None
    ) -> None:
        if depth_left is None:
            depth_left = self._config["ask_depth"]
        assert depth_left is not None
        if depth_left < 0:
            return
        self.clear_timer(clear_preserved=False)
        if chat_context is None:
            chat_context = self.chat_context
        self._ask_context.extend(messages_list)
        if not self._check_task_validity():
            return
        response = self.to_llm_method(chat_context, self._ask_context)
        if not self._check_task_validity():
            return
        choice_code, content = parse_response(self, response)
        if not self._check_task_validity():
            return
        choice_execute(ai_care=self, choice_code=choice_code, content=content, depth_left=depth_left)

    def set_cyclic_detection(
        self,
        detectors: list[str],
        interval_gen: Generator[float, None, None],
        cancel_after_trigger_ask: bool = True
    ) -> None:
        def callback():
            for detector in detectors:
                self.release_detector(detector)
            try:
                self.set_timer(
                    interval=next(interval_gen),
                    function=callback,
                    preserve=not cancel_after_trigger_ask,
                )
            except StopIteration:
                pass
        try:
            self.set_timer(
                interval=next(interval_gen),
                function=callback,
                preserve=not cancel_after_trigger_ask,
            )
        except StopIteration:
            pass

    def trigger(
        self,
        messages_list: list[AICareContext] | None = None,
        tag: str | None = None,
        chat_context: ChatContext | None = None,
        depth_left: int = 1
    ) -> None:
        if messages_list:
            self.ask(chat_context=chat_context, messages_list=messages_list, depth_left=depth_left)
        if tag:
            for detector in self._tags[tag]:
                self.release_detector(detector.name)

    def register_detector(self, detector: Detector) -> None:
        if detector.name in self.detectors:
            raise ValueError("A detector with the same name already exists.")
        detector.ai_care = self
        self.detectors[detector.name] = detector
        tag = detector.tag
        if tag not in self._tags:
            self._tags[tag] = []
        self._tags[tag].append(detector)

    def release_detector(self, name: str | list[str]) -> None:
        if isinstance(name, list):
            names = name
        else:
            names = [name]
        for name in names:
            if name not in self.detectors:
                raise ValueError(f"There is no detector named {name}.")
            detector = self.detectors[name]
            ai_care_thread = AICareThread(target=detector.release)
            ai_care_thread._task_num = self._get_task_num()
            ai_care_thread.start()

    def register_sensor(self, name: str, function: Callable[[], Any], annotation: str) -> None:
        if name in self.sensors:
            raise ValueError(f"The sensor named {name} has already been registered.")
        self.sensors[name] = {"name": name, "function": function, "annotation": annotation}

    def get_sensor_data(self, name: str) -> Any:
        if name not in self.sensors:
            raise ValueError("No sensor named {name}.")
        sensor_function = self.sensors[name]["function"]
        data = sensor_function()
        return data


class Detector(metaclass=ABCMeta):
    def __init__(self, name: str, annotation: str, tag: str = '') -> None:
        self.name = name
        self.annotation = annotation
        self.tag = tag
        self.ai_care: AICare | None = None

    @abstractmethod
    def detect(self) -> bool:
        """Here, execute the detection logic and return whether it has triggered ai_care."""
        ...

    def release(self) -> None:
        self.detect()


class AICareContext(TypedDict):
    role: Literal["ai_care", "assistant"]
    content: str


class AICareTimer(threading.Timer):
    def __init__(self, *args, preserve: bool = False, daemon: bool = True, **kwargs):
        super().__init__(*args, **kwargs)
        self._preserve_: bool = preserve
        self._task_num: int = 0
        self.daemon = daemon


class AICareThread(threading.Thread):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._task_num: int = 0



import pytest
import time
from unittest.mock import Mock

from ai_care import AICare
from ai_care.abilities import Ability, _ability


@pytest.fixture
def ai_care():
    ai_care = AICare()
    yield ai_care
    ai_care.clear_timer(clear_preserved=True, default_task_num_authority_external="Highest")

def test_auto_register_abilities():
    class FakeAbility(Ability):
        @_ability(
            description="Fake_ability_description"
        )
        def fake_ability(self) -> str:
            return "Fake ability"

        def not_ability(self) -> str:
            return "Not ability"

    fake_ability = FakeAbility(Mock())

    abilities_list = list(fake_ability.abilities.keys())
    
    assert "fake_ability" in abilities_list
    assert "not_ability" not in abilities_list

    result = []
    for name, ability in fake_ability.abilities.items():
        if name == "fake_ability":
            assert ability._ability_description_ == "Fake_ability_description"
            result.append(ability())

    assert result == ["Fake ability"]

def test_abilities(ai_care: AICare):
    # Setup
    response = (x for x in ['AA0', '0030', '3:{"delay": 0.2, ', '"message": ', '"hi"}'])
    ai_care.set_config(key="delay", value=0.1)
    to_user_method = Mock()
    to_llm_method = Mock()
    to_llm_method.return_value = response
    ai_care.register_to_llm_method(to_llm_method)
    ai_care.register_to_user_method(to_user_method)
    
    # Action
    ai_care.chat_update(chat_context=[])
    
    # Assert
    time.sleep(0.2)
    assert to_llm_method.called
    time.sleep(0.2)
    assert to_user_method.called
    called_args, _ = to_user_method.call_args
    assert ''.join(x for x in called_args[0]) == "hi"

def test_detect_env(ai_care: AICare):
    # Setup
    ai_care.ask = Mock()
    mock_sensor = Mock()
    mock_sensor.return_value = "mock_sensor value"
    ai_care.sensors = {"mock_sensor": {"name": "mock_sensor", "function": mock_sensor, "annotation": "mock_sensor_annotation"}}

    # Action
    ai_care.ability.abilities["detect_env"](delay=0.1, sensors=["mock_sensor"], _depth_left=1)
    time.sleep(0.2)

    # Assert
    assert ai_care.ask.called
    _, called_kwargs = ai_care.ask.call_args
    assert "The results of the sensor are as follows: {'mock_sensor': 'mock_sensor value'}" in called_kwargs["messages_list"][0]["content"]

    # Action
    ai_care.ability.abilities["detect_env"](delay=0.1, sensors=["mock_sensor_not_existed"], _depth_left=1)
    time.sleep(0.2)

    # Assert
    _, called_kwargs = ai_care.ask.call_args
    assert "There are no {'mock_sensor_not_existed'} sensor." in called_kwargs["messages_list"][0]["content"]

def test_release_detector(ai_care: AICare):
    # Setup
    ai_care.ask = Mock()
    mock_detector = Mock()
    ai_care.detectors = {"mock_detector": mock_detector}
    mock_release_detector = Mock()
    ai_care.release_detector = mock_release_detector

    # Action
    ai_care.ability.abilities["release_detector"](delay=0.1, detectors=["mock_detector_not_existed"], _depth_left=1)

    # Assert
    assert ai_care.ask.called
    _, called_kwargs = ai_care.ask.call_args
    assert "There are no {'mock_detector_not_existed'} detector." in called_kwargs["messages_list"][0]["content"]
    
    # Action
    ai_care.ability.abilities["release_detector"](delay=0.1, detectors=["mock_detector"], _depth_left=1)
    time.sleep(0.2)

    # Assert
    mock_release_detector.assert_called_with(["mock_detector"])

def test_ask_later(ai_care: AICare):
    # Setup
    ai_care.ask = Mock()

    # Action
    ai_care._ask_later_count_left = 0
    ai_care.ability.abilities["ask_later"](delay=0.1, _depth_left=1)
    time.sleep(0.2)

    # Assert
    assert not ai_care.ask.called
    
    # Action
    ai_care._ask_later_count_left = 1
    ai_care.ability.abilities["ask_later"](delay=0.1, _depth_left=1)
    time.sleep(0.2)

    # Assert
    _, called_kwargs = ai_care.ask.call_args
    assert "0.1 seconds have passed. Please make a choice." in called_kwargs["messages_list"][0]["content"]
    
    # Action
    ai_care.ability.abilities["ask_later"](delay=0.1, _depth_left=1)
    time.sleep(0.2)

    # Assert
    assert ai_care.ask.call_count == 1

def test_cyclic_detection(ai_care: AICare):
    # Setup
    ai_care.ask = Mock()
    mock_detector = Mock()
    ai_care.detectors = {"mock_detector": mock_detector}
    mock_release_detector = Mock()
    ai_care.release_detector = mock_release_detector

    # Action
    ai_care.ability.abilities["cyclic_detection"](interval=0.1, detectors=["mock_detector_not_existed"], _depth_left=1)
    time.sleep(0.2)

    # Assert
    _, called_kwargs = ai_care.ask.call_args
    assert "There are no {'mock_detector_not_existed'} detector." in called_kwargs["messages_list"][0]["content"]
    assert not mock_release_detector.called
    
    # Action
    ai_care.ability.abilities["cyclic_detection"](interval=0.1, detectors=["mock_detector"], _depth_left=1)
    time.sleep(0.35)
    ai_care.clear_timer()
    time.sleep(0.3)

    # Assert
    assert mock_release_detector.call_count == 3



from __future__ import annotations
import logging
from typing import TYPE_CHECKING, Generator, cast

from .abilities import Choice


if TYPE_CHECKING:
    from .ai_care import AICare


logger = logging.getLogger("ai_care")


def parse_response(
    ai_care: AICare,
    response: str | Generator[str, None, None],
) -> tuple[str, str | Generator[str, None, None]]:
    choice_code = ""
    content = ""
    if isinstance(response, str):
        ai_care._stream_mode = False
        prefix, content = _extract_info(response)
        if not all(prefix):
            return '00', ''
        choice_code = prefix[2]
        check_valid = prefix[3]
        if choice_code == check_valid:
            ai_care._valid_msg_count += 1
        else:
            ai_care._invalid_msg_count += 1
        assert isinstance(choice_code, str)
        return choice_code, content or ""
    elif isinstance(response, Generator):
        ai_care._stream_mode = True
        buffer = ""
        first_item = ""
        chunk_list = []
        found_choice = False
        prefix = (None, None, None, None)
        choice = Choice.ERROR
        for chunk in response:
            buffer += chunk
            if not found_choice:
                prefix, content = _extract_info(buffer)
            if all(prefix) and len(buffer) >= 9 and not found_choice:
                choice_code = prefix[2]
                check_valid = prefix[3]
                if choice_code == check_valid:
                    ai_care._valid_msg_count += 1
                else:
                    ai_care._invalid_msg_count += 1
                try:
                    choice = Choice(choice_code)
                except ValueError as e:
                    logger.warning(f"Invalid choice {choice_code}. Error: {e}")
                    assert isinstance(choice_code, str)
                    return choice_code, ''
                found_choice = True
                if choice == Choice.SPEAK_NOW:
                    first_item = buffer[9:]
                    break
                chunk_list.append(buffer[9:])
                continue
            if found_choice:
                chunk_list.append(chunk)
        if found_choice is False:
            return '00', ''
        prefix = cast(tuple[str, str, str, str], prefix)
        if choice == Choice.SPEAK_NOW:
            def response_content_gen():
                gen_content_record = []
                yield first_item
                gen_content_record.append(first_item)
                for item in response:
                    yield item
                    gen_content_record.append(item)
                ai_care._ask_context.append(
                    {
                        "role": "ai_care",
                        "content": ''.join(gen_content_record),
                    }
                )
            return choice.value, response_content_gen()
        else:
            return choice.value, ''.join(chunk_list)
    else:
        assert False, "The response must be str or a generator."

def _extract_info(s: str) -> tuple[tuple[str | None, str | None, str | None, str | None], str | None]:
    """Parse a given string.

    It extracts two-letter groups from the first eight characters
    and the remainder of the string after the ninth character. If a group is not complete (less than two characters),
    it is set to None.

    Parameters:
    s (str): The string to be parsed.

    Returns:
    tuple: A tuple containing a tuple of four two-letter groups and the remainder of the string.
    """
    # Extract the first eight characters and split them into four groups
    groups = [s[i:i+2] if len(s[i:i+2]) == 2 else None for i in range(0, 8, 2)]

    # Extract the remainder of the string after the ninth character
    remainder = s[9:] if len(s) > 9 else None

    return (tuple(groups), remainder)



from .ai_care import AICare, Detector, AICareContext
from ._version import __title__, __version__


__all__ = [
    "__title__",
    "__version__",
    "AICare",
    "Detector",
    "AICareContext",
]



import pytest
import time
from unittest.mock import Mock, patch

from ai_care import AICare, Detector


@pytest.fixture
def ai_care():
    ai_care = AICare()
    yield ai_care
    ai_care.clear_timer(clear_preserved=True, default_task_num_authority_external="Highest")

def test_set_timer(ai_care: AICare):
    # Setup
    callback = Mock()
    callback_args = ("arg1", "arg2")
    callback_kwargs = {"kwarg1": "value1", "kwarg2": "value2"}

    # Action
    timer_id = ai_care.set_timer(interval=0.2, function=callback, args=callback_args, kwargs=callback_kwargs)

    # Assert
    assert timer_id in ai_care.timers
    time.sleep(0.1)
    assert not callback.called
    time.sleep(0.2)
    assert callback.called
    assert callback.call_count == 1
    callback.assert_called_with(*callback_args, **callback_kwargs)
    assert timer_id not in ai_care.timers

def test_timer_cancel(ai_care: AICare):
    # Setup
    callback = Mock()
    callback_args = ("arg1", "arg2")
    callback_kwargs = {"kwarg1": "value1", "kwarg2": "value2"}

    # Action
    timer_id = ai_care.set_timer(interval=0.2, function=callback, args=callback_args, kwargs=callback_kwargs)

    # Assert
    assert timer_id in ai_care.timers

    # Action
    time.sleep(0.1)
    ai_care.timer_cancel(timer_id)

    # Assert
    assert timer_id not in ai_care.timers
    time.sleep(0.2)
    assert not callback.called

def test_set_guide(ai_care: AICare):
    # Action
    ai_care.set_guide("You are a considerate person who cars about others.")

    # Assert
    assert ai_care.guide == "You are a considerate person who cars about others."

def test_task_num(ai_care: AICare):
    # Setup
    ai_care._task_num = 5
    
    flag = 0
    def callback(ai_care: AICare) -> None:
        nonlocal flag
        task_num = ai_care._get_task_num()
        flag = task_num
    
    # Action
    ai_care.set_timer(interval=0.1, function=callback, args=(ai_care,), task_num=3)
    time.sleep(0.2)

    # Assert
    assert ai_care._get_task_num() == 5
    assert flag == 3

    # Action
    ai_care.set_timer(interval=0.1, function=callback, args=(ai_care,))
    time.sleep(0.2)

    # Assert
    assert flag == 5

def test_clear_timer(ai_care: AICare):
    # Setup
    callback1 = Mock()
    callback2 = Mock()
    callback3 = Mock()
    callback4 = Mock()

    # Action
    ai_care._task_num = 1
    ai_care.set_timer(interval=0.1, function=callback1)
    ai_care._task_num = 2
    ai_care.set_timer(interval=0.1, function=callback2)
    ai_care.clear_timer(task_num_authority=1)
    time.sleep(0.2)

    # Assert
    assert not callback1.called
    assert callback2.called

    # Action
    def cancel_timer(ai_care: AICare):
        ai_care.clear_timer()
    ai_care._task_num = 1
    ai_care.set_timer(interval=0.2, function=callback3)
    ai_care.set_timer(interval=0.1, function=cancel_timer, kwargs={"ai_care": ai_care})
    ai_care._task_num = 2
    ai_care.set_timer(interval=0.2, function=callback4)
    time.sleep(0.3)

    # Assert
    assert not callback3.called
    assert callback4.called

def test_insert_chat_interval(ai_care: AICare):
    # Action
    for i in range(5):
        ai_care._insert_chat_interval(i)

    # Assert
    assert [int(i) for i in ai_care._chat_intervals] == [0, 1, 2, 3, 4]
    
    # Action
    for i in range(1, 31):
        ai_care._insert_chat_interval(i)

    # Assert
    assert [int(i) for i in ai_care._chat_intervals] == list(range(11, 31))

def test_chat_update(ai_care: AICare):
    # Setup
    mock_ask = Mock()
    
    # Action
    ai_care.set_config(key="delay", value=0.1)
    ai_care.set_config(key="ask_later_count_limit", value=1)
    ai_care.set_timer(interval=0.1, function=Mock())
    ai_care.ask = mock_ask
    ai_care.chat_update(chat_context=["chat_context"])
    time.sleep(0.2)

    # Assert
    assert ai_care._task_num == 2
    assert ai_care._chat_intervals == []
    assert ai_care.chat_context == ["chat_context"]
    assert ai_care._ask_later_count_left == 1
    assert ai_care.timers == {}
    assert mock_ask.called

def test_reset(ai_care: AICare):
    # Setup
    mock_task = Mock()
    ai_care._valid_msg_count = 10
    ai_care._invalid_msg_count = 10
    ai_care._last_chat_time = 10
    ai_care._chat_intervals = [10,10]
    ai_care.set_timer(interval=0.1, function=mock_task, preserve=True)

    # Action
    ai_care.reset()
    time.sleep(0.2)
    
    # Assert
    assert ai_care._valid_msg_count == 0
    assert ai_care._invalid_msg_count == 0
    assert ai_care._last_chat_time is None
    assert ai_care._chat_intervals == []
    assert not mock_task.called

def test_check_task_validity(ai_care: AICare):
    # Setup
    tasks_validity = []
    def check_validity(ai_care: AICare, validity_list: list):
        validity_list.append(ai_care._check_task_validity())

    # Action
    ai_care.set_timer(interval=0.1, function=check_validity, args=(ai_care, tasks_validity))
    ai_care.cancel_current_task()
    ai_care.set_timer(interval=0.1, function=check_validity, args=(ai_care, tasks_validity))
    time.sleep(0.2)

    # Assert
    assert tasks_validity == [False, True]

def test_ask(ai_care: AICare):
    # Setup
    mock_to_llm_method = Mock()
    mock_to_llm_method.return_value = "AA000101:"

    ai_care._to_llm_method = mock_to_llm_method

    # Action
    ai_care.ask(messages_list=[], depth_left=-1)

    # Assert
    assert not mock_to_llm_method.called

    # Action
    with patch('ai_care.ai_care.parse_response') as mock_pares_response, \
        patch('ai_care.ai_care.choice_execute') as mock_choice_execute:
        mock_pares_response.return_value = ("01", "")
        ai_care.ask(messages_list=[])

    # Assert
    assert mock_to_llm_method.called
    assert mock_pares_response.called
    assert mock_choice_execute.called
    mock_choice_execute.assert_called_with(
        ai_care=ai_care,
        choice_code="01",
        content="",
        depth_left=ai_care._config["ask_depth"]
    )

def test_trigger(ai_care: AICare):
    # Setup
    mock_ask = Mock()
    mock_release_detector = Mock()
    mock_detector = Mock()
    mock_detector.name = "detector_name"
    ai_care._tags = {"detector_tag": [mock_detector]}
    ai_care.ask = mock_ask
    ai_care.release_detector = mock_release_detector
    
    # Action
    ai_care.trigger(messages_list=[{"role": "ai_care", "content": "content"}])
    
    # Assert
    mock_ask.assert_called_with(
        chat_context=None,
        messages_list=[{"role": "ai_care", "content": "content"}],
        depth_left=1
    )
    assert not mock_release_detector.called

    # Action
    ai_care.trigger(tag="detector_tag")
    mock_release_detector.assert_called_with("detector_name")

def test_register_detector(ai_care: AICare):
    # Setup
    mock_detector = Mock()
    mock_detector.name = "mock_detector_name"
    mock_detector.tag = "mock_detector_tag"

    # Action
    ai_care.register_detector(mock_detector)

    # Assert
    assert mock_detector.ai_care == ai_care
    assert ai_care.detectors["mock_detector_name"] == mock_detector
    assert ai_care._tags["mock_detector_tag"] == [mock_detector]

def test_release_detector(ai_care: AICare):
    # Setup
    det_ann_called_list = []
    class TestDetector(Detector):
        def detect(self) -> bool:
            det_ann_called_list.append(self.annotation)
            return True
    test_detector1 = TestDetector(name="test_detector1", annotation="test_detector1_annotation")
    test_detector2 = TestDetector(name="test_detector2", annotation="test_detector2_annotation")
    test_detector3 = TestDetector(name="test_detector3", annotation="test_detector3_annotation")

    # Action
    ai_care.register_detector(test_detector1)
    ai_care.register_detector(test_detector2)
    ai_care.register_detector(test_detector3)
    ai_care.release_detector("test_detector1")
    ai_care.release_detector(["test_detector2", "test_detector3"])
    time.sleep(0.1)

    # Assert
    assert set(det_ann_called_list) == {"test_detector1_annotation", "test_detector2_annotation", "test_detector3_annotation"}

def test_register_sensor(ai_care: AICare):
    # Setup
    mock_sensor = Mock()
    mock_sensor.return_value = 10

    # Action
    ai_care.register_sensor(name="mock_sensor", function=mock_sensor, annotation="mock_sensor_annotation")

    # Assert
    assert ai_care.sensors["mock_sensor"] == {"name": "mock_sensor", "function": mock_sensor, "annotation": "mock_sensor_annotation"}

def test_get_sensor_data(ai_care: AICare):
    # Setup
    mock_sensor = Mock()
    mock_sensor.return_value = 10

    # Action
    ai_care.register_sensor(name="mock_sensor", function=mock_sensor, annotation="mock_sensor_annotation")

    # Assert
    assert ai_care.get_sensor_data("mock_sensor") == 10

def test_set_cyclic_detection(ai_care: AICare):
    # Setup
    mock_release_detector_method_1 = Mock()
    ai_care.release_detector = mock_release_detector_method_1

    # Action
    ai_care.set_cyclic_detection(detectors=["fake_release_detector"], interval_gen=(0.1 for _ in range(3)))
    time.sleep(0.6)

    # Assert
    assert mock_release_detector_method_1.called
    assert mock_release_detector_method_1.call_count == 3
    
    # Setup
    mock_release_detector_method_2 = Mock()
    ai_care.release_detector = mock_release_detector_method_2
    ai_care.to_llm_method = Mock()

    # Action
    with patch('ai_care.ai_care.parse_response') as mock_pares_response, patch('ai_care.ai_care.choice_execute'):
        mock_pares_response.return_value = ("01", "fake content")
        ai_care.set_cyclic_detection(detectors=["fake_release_detector"], interval_gen=(x for x in [0.1, 0.1, 0.2, 0.1, 0.1]))
        time.sleep(0.3)
        ai_care.trigger(messages_list=[{"role": "ai_care", "content": ""}])
        time.sleep(0.4)

    # Assert
    assert mock_release_detector_method_2.call_count == 2



from __future__ import annotations
import json
import logging
from typing import TYPE_CHECKING, Generator

from .abilities import Choice


logger = logging.getLogger("ai_care")


if TYPE_CHECKING:
    from .ai_care import AICare


def choice_execute(ai_care: AICare, choice_code: str, content: str | Generator[str, None, None], depth_left: int) -> None:
    try:
        choice = Choice(choice_code)
    except ValueError as e:
        logger.warning(f"Invalid choice {choice_code}.")
        ai_care.ask(
            messages_list=[
                {
                    "role": "assistant",
                    "content": f"AA00{choice_code}{choice_code}:{content}",
                },
                {
                    "role": "ai_care",
                    "content": f"Your choice code {choice_code} is not correct. Please make a correct choice again.",
                },
            ],
            depth_left = depth_left - 1,
        )
        return

    if choice == Choice.ERROR:
        ai_care.ask(messages_list=[], depth_left = depth_left - 1)
        return
    logger.info(f"Choice: {choice.name}")

    if isinstance(content, str):
        ai_care._ask_context.append(
            {
                "role": "assistant",
                "content": f"AA00{choice_code}{choice_code}:{content}",
            }
        )
    elif isinstance(content, Generator):
        # This case has been handled in parse_response.
        pass
    else:
        assert False

    ability_method = ai_care.ability.abilities[choice.name.lower()]
    
    if choice == Choice.STAY_SILENT:
        ability_method()
        return
    
    if choice == Choice.SPEAK_NOW:
        params = {"content": content}
    else:
        assert isinstance(content, str)
        try:
            params = json.loads(content)
        except json.JSONDecodeError as e:
            logger.warning(f"Failed to correctly parse the parameter. Parameter json string: {content}. Error: {e}.")
            ai_care.ask(
                messages_list=[
                    {
                        "role": "ai_care",
                        "content": (
                            "Failed to correctly parse the parameter. "
                            "Please send the correct parameters in JSON format, "
                            "or make a choice again."
                        ),
                    },
                ],
                depth_left = depth_left - 1,
            )
            return
        if not isinstance(params, dict):
            ai_care.ask(
                messages_list=[
                    {
                        "role": "ai_care",
                        "content": (
                            "The parameters should be a dictionary in JSON format."
                            "Please send the correct parameters in JSON format, or make a choice again."
                        ),
                    },
                ],
                depth_left = depth_left - 1,
            )
            return

    ability_params = {}
    for param in ability_method._ability_parameters_:
        default_value = param["default_value"]
        if param["required"] is False:
            ability_params[param["name"]] = default_value
    ability_params.update(params)
    if getattr(ability_method, "_auto_depth_", False):
        ability_params[ability_method._depth_param_name_] = depth_left

    needed_params_set = {param["name"] for param in ability_method._ability_parameters_}
    missing_params_set = needed_params_set - set(ability_params.keys())

    if missing_params_set:
        ai_care.ask(
            messages_list=[
                {
                    "role": "ai_care",
                    "content": (
                        f"You did not provide the following parameters: {str(missing_params_set)} ."
                        "Please send the correct parameters in JSON format, or make a choice again."
                    ),
                },
            ],
            depth_left = depth_left - 1
        )
        return

    logger.info(f"Choice parameters: {str(ability_params)}")
    ability_method(**ability_params)



from unittest.mock import Mock

from ai_care.choice_execute import choice_execute


def test_choice_execute_choice_error():
    # Setup
    mock_ai_care = Mock()

    # Action
    choice_execute(ai_care=mock_ai_care, choice_code="EE", content="", depth_left=1)

    # Assert
    _, called_kwargs = mock_ai_care.ask.call_args
    assert "Your choice code EE is not correct." in called_kwargs["messages_list"][1]["content"]

def test_choice_execute_parse_json_error():
    # Setup
    mock_ai_care = Mock()
    mock_ai_care.ability.abilities = {
        "stay_silent": Mock(),
        "speak_now": Mock(),
        "speak_after": Mock()
    }

    # Action
    choice_execute(ai_care=mock_ai_care, choice_code="03", content="{wrong json string", depth_left=1)

    # Assert
    _, called_kwargs = mock_ai_care.ask.call_args
    assert "Failed to correctly parse the parameter." in called_kwargs["messages_list"][0]["content"]

def test_choice_execute_param_not_dict_error():
    # Setup
    mock_ai_care = Mock()
    mock_ai_care.ability.abilities = {
        "stay_silent": Mock(),
        "speak_now": Mock(),
        "speak_after": Mock()
    }

    # Action
    choice_execute(ai_care=mock_ai_care, choice_code="03", content='[{"role": "ai_care"}]', depth_left=1)

    # Assert
    _, called_kwargs = mock_ai_care.ask.call_args
    assert "The parameters should be a dictionary in JSON format." in called_kwargs["messages_list"][0]["content"]

def test_choice_execute_error():
    # Setup
    mock_ai_care = Mock()

    # Action
    choice_execute(ai_care=mock_ai_care, choice_code="00", content="", depth_left=1)

    # Assert
    mock_ai_care.ask.assert_called_with(messages_list=[], depth_left=0)

def test_choice_execute_stay_silent():
    # Setup
    mock_ai_care = Mock()
    mock_stay_silent_method = Mock()
    mock_speak_now_method = Mock()
    mock_speak_after_method = Mock()
    mock_ai_care.ability.abilities = {
        "stay_silent": mock_stay_silent_method,
        "speak_now": mock_speak_now_method,
        "speak_after": mock_speak_after_method
    }
    depth_left = 3
    choice_code = "01"
    content = "content"

    # Action
    choice_execute(ai_care=mock_ai_care, choice_code=choice_code, content=content, depth_left=depth_left)

    # Assert
    assert mock_stay_silent_method.called
    assert not mock_speak_now_method.called
    assert not mock_speak_after_method.called

def test_choice_execute_speak_now():
    # Setup
    mock_ai_care = Mock()
    mock_stay_silent_method = Mock()
    mock_speak_now_method = Mock()
    mock_speak_after_method = Mock()
    mock_ai_care.ability.abilities = {
        "stay_silent": mock_stay_silent_method,
        "speak_now": mock_speak_now_method,
        "speak_after": mock_speak_after_method
    }
    mock_speak_now_method._ability_parameters_ = [
        {
            "name": "content",
            "description": "The content you want to say",
            "param_type": "string",
            "required": True,
            "default_value": "",
        },
    ]
    mock_speak_now_method._auto_depth_ = False
    depth_left = 3
    choice_code = "02"
    content = "content"

    # Action
    choice_execute(ai_care=mock_ai_care, choice_code=choice_code, content=content, depth_left=depth_left)

    # Assert
    assert not mock_stay_silent_method.called
    assert not mock_speak_after_method.called
    assert mock_speak_now_method.called
    mock_speak_now_method.assert_called_with(content="content")

def test_choice_execute_speak_after():
    # Setup
    mock_ai_care = Mock()
    mock_stay_silent_method = Mock()
    mock_speak_now_method = Mock()
    mock_speak_after_method = Mock()
    mock_ai_care.ability.abilities = {
        "stay_silent": mock_stay_silent_method,
        "speak_now": mock_speak_now_method,
        "speak_after": mock_speak_after_method
    }
    mock_speak_after_method._ability_parameters_ = [
        {
            "name": "delay",
            "description": "",
            "param_type": "string",
            "required": True,
            "default_value": "",
        },
        {
            "name": "message",
            "description": "",
            "param_type": "string",
            "required": True,
            "default_value": "",
        },
    ]
    mock_speak_after_method._auto_depth_ = False
    depth_left = 3
    choice_code = "03"
    content = '{"delay": 1, "message": "message content"}'

    # Action
    choice_execute(ai_care=mock_ai_care, choice_code=choice_code, content=content, depth_left=depth_left)

    # Assert
    assert not mock_stay_silent_method.called
    assert not mock_speak_now_method.called
    assert mock_speak_after_method.called
    mock_speak_after_method.assert_called_with(delay=1, message="message content")



from __future__ import annotations
import textwrap
import time
from typing import TYPE_CHECKING

from .abilities import Choice


if TYPE_CHECKING:
    from .ai_care import AICare


def render_basic_prompt(
    ai_care: AICare,
    inactive_abilities_list: list[Choice] | None = None,
    inactive_sensors_list: list[str] | None = None,
    inactive_detectors_list: list[str] | None = None,
) -> str:
    inactive_abilities_set = set() if inactive_abilities_list is None else set(inactive_abilities_list)
    inactive_sensors_set = set() if inactive_sensors_list is None else set(inactive_sensors_list)
    inactive_detectors_set = set() if inactive_detectors_list is None else set(inactive_detectors_list)
    if ai_care._ask_later_count_left <= 0:
        inactive_abilities_set.add(Choice.ASK_LATER)
    
    abilities_dict = ai_care.ability.abilities
    
    intervals_info = (
        f"""The intervals of the last {len(ai_care._chat_intervals)} times the user conversed with you are recorded in the following list (unit in seconds):
        {str(ai_care._chat_intervals)}
        """ if ai_care._chat_intervals else ""
    ) + (
        f"""It has been {time.monotonic() - ai_care._last_chat_time} seconds since the last time the user spoke with you."""
        if ai_care._last_chat_time is not None else ""
    )

    sorted_abilities = sorted(abilities_dict.values(), key=lambda x: Choice[x.__name__.upper()].value)
    abilities_info = ''.join(_render_ability_description(ability_method).lstrip()
        for ability_method in sorted_abilities
        if ability_method not in inactive_abilities_set
    ).replace('\n', '\n        ')

    prompt = textwrap.dedent(
        f"""
        I am a program, and my name is Aicarey.
        You can see your conversation history with the user in the previous messages.
        This message is sent by me. Please continue to focus on the user and do not attempt to converse with me.
        You should seriously judge which choice you should make based on the content of your conversation with the user.
        For example, if you are in a question-and-answer mode with the user,
        then you may not need to speak when the user doesn't ask a question.
        However, if you are chatting with the user like a friend,
        then you may need to proactively continue the conversation like a friend when the user doesn't speak.

        When replying to this message, you must follow the rules below:
        ========================RESPONSE RULES=======================
        1. Start with eight characters followed by an English colon.
        The first two characters of these eight must be 'AA', the third and fourth must be '00',
        the fifth and sixth are the code of your choice.
        The seventh and eighth characters should repeat the fifth and sixth characters.
        2. After the colon is the content corresponding to the chosen option.
        If it involves a function call, this part of the content must be in the format of a JSON string.
        =============================================================

        Here are the choices you can make:
        ===========================CHOICES===========================
        {abilities_info}
        =============================================================
        You must choose one of the options provided above as your reply.
        
        Response Examples:
            If you want to remain silent: AA000101:
            If you want to say to the user: AA000202: [Your content here]
            If you decide to ask the user what they are doing if they haven't spoken to you in a minute: AA000303: {{"delay":60, "message":"What are you doing?"}}

        ===========================SENSORS===========================
        Sensors list:
        {
            str(
                [
                    {"sensor_name": sensor["name"], "sensor_description": sensor["annotation"]}
                    for sensor in ai_care.sensors.values()
                    if sensor["name"] not in inactive_sensors_set
                ]
            )
        }
        =============================================================
        
        ==========================DETECTORS==========================
        Detectors list:
        {
            str(
                [
                    {"detector_name": detector.name, "detector_description": detector.annotation}
                    for detector in ai_care.detectors.values()
                    if detector.name not in inactive_detectors_set
                ]
            )
        }
        =============================================================
        
        ============================FACTS============================
        {intervals_info}
        
        {ai_care.guide}
        =============================================================
        """
    )
    return prompt

def _render_ability_description(ability_method) -> str:
    choice = Choice[ability_method.__name__.upper()]
    content_string = "The content you want to say\n\n" if choice == Choice.SPEAK_NOW \
        else "Provide the values of the following parameters in JSON format\n" if ability_method._ability_parameters_ \
        else "Empty\n\n"
    if choice not in {Choice.SPEAK_NOW, Choice.STAY_SILENT}:
        for param_dict in ability_method._ability_parameters_:
            content_string += f"    Parameter name: {param_dict['name']}\n"
            content_string += f"    Parameter description: {param_dict['description']}\n"
            content_string += f"    Parameter type: {param_dict['param_type']}\n"
            if param_dict["required"] is False:
                content_string += f"    This parameter is optional, with a default value: {param_dict['default_value']}\n\n"
            else:
                content_string += f"    This parameter must be provided\n\n"
    prompt = (
        "\n"
        f"Choice code: {choice.value}\n"
        f"Choice description: {ability_method._ability_description_}\n"
        f"Choice content: {content_string}"
    )
    return prompt
